/*
 * IRRC - client (WinLIRC compatible, Windows-specific)
 * Copyright (c) 2003 Jan Wassenberg
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Contact info:
 *   Jan.Wassenberg@stud.uni-karlsruhe.de
 *   http://www.stud.uni-karlsruhe.de/~urkt/
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <sys/stat.h>
#include <string.h>

#include "posix.h"
#include "misc.h"

#include "sysdep/sysdep.h"



#include "sysdep/win/win_internal.h"

//#pragma warning(disable: 4201)	/* MS extensions */
//#include <mmsystem.h>	/* waveOut?etVolume */

// easier than including header and fixing conflicts
#ifdef __cplusplus
extern "C" {
#endif
uint __stdcall waveOutGetVolume(void* hwo, unsigned long* pdwVolume);
uint __stdcall waveOutSetVolume(void* hwo, unsigned long dwVolume);
#ifdef __cplusplus
}
#endif

static const int ERR_BUF_SIZE = 1000;
static char err[ERR_BUF_SIZE];
static char* err_pos = err;

static void err_append(const char* msg, ...)
{
	va_list ap;
	va_start(ap, msg);
	err_pos += vsprintf(err_pos, msg, ap);
	va_end(ap);
}

static bool err_occurred()
{
	return err_pos != err;
}


#include "irrc.h"



/*****************************************************************************
 * config (key definitions, app / command binding)
 ****************************************************************************/

#define MAX_KEYS 128

static char* key_strs[MAX_KEYS];


#define MAX_CMDS 32

static struct APP
{
	HWND hwnd;
	char* class_name;
	char* exe;

	int num_cmds;
	char* rc_keys[MAX_CMDS];	/* string identifying remote control key */
	char keys[MAX_CMDS];		/* corresponding char sent to the app */
	bool allow_rep[MAX_CMDS];
	char modifier[MAX_CMDS];
	int reps[MAX_CMDS];
}
apps[10];

static struct APP* active = &apps[1];


#define NUM_VK (sizeof(vk)/sizeof(vk[0]))

/* don't use X macros: there aren't many, and some strings / VK names are different */

static char* vk_str[] =
{
	"UP", "DOWN", "LEFT", "RIGHT",
	"PLUS", "MINUS", "RETURN", "SPACE",
	"PAGEUP", "PAGEDOWN",
	"SHIFT", "CTRL", "ALT"
};

static char vk[] =
{
	VK_UP, VK_DOWN, VK_LEFT, VK_RIGHT,
	VK_ADD, VK_SUBTRACT, VK_RETURN, VK_SPACE,
	VK_NEXT, VK_PRIOR,
	VK_SHIFT, VK_CONTROL, VK_MENU
};


static char buf[16384];
char* p1 = &buf[0], * p2 = &buf[8192];

	struct APP* app = 0;
	int line = 0;
	int cmds = 0;



static void key_def(const char* s)
{
	uint code;

	if(sscanf(s, "0x%x %[^\t\n\r]", &code, p1) != 2)
	{
		err_append("  ignoring line %d (expected hex code and key name, separated by whitespace)\n", line);
		return;
	}

	if(code >= MAX_KEYS)
	{
		err_append("  ignoring line %d (code 0x%x > maximum allowed value of 0x%x)\n", line, code, MAX_KEYS);
		return;
	}

	if(key_strs[code] != 0)
	{
		err_append("  ignoring line %d (code 0x%02x is already bound to \"%s\")\n", line, code, key_strs[code]);
		return;
	}

	key_strs[code] = p1;
	p1 += strlen(p1)+1;
}


static void app_header(char* s)
{
	int num = s[0] - '0';
	app = &apps[num];

	if(sscanf(s+1, " %[^\t] %[^\t\n\r]", p1, p2) != 2)
	{
		err_append("  ignoring line %d (expected window class name and EXE path, delimited by tab[s])\n", line);
		return;
	}

	if(app->exe)
	{
		err_append("  ignoring line %d (number %d is already assigned to app \"%s\")\n", line, num, app->exe);
		return;
	}

	if(stat(p2, (struct stat*)s) != 0)	/* don't care about stat struct, can trash s */
	{
		err_append("  ignoring line %d (file \"%s\" not found)\n", line, p2);
		return;
	}

	app->class_name = p1;
	app->exe = p2;

	p1 += strlen(p1)+1;
	p2 += strlen(p2)+1;
}


static void key_binding(char* s)
{
	/* check if auto-repeatable */
	bool allow_rep = false;
	if(s[0] == '*')
	{
		s[0] = ' ';
		allow_rep = true;
	}

	if(sscanf(s, " %[^\t] %[^\t\n\r]", p1, p2) != 2)
	{
		err_append("  ignoring line %d (expected RC key name string and target key, delimited by tab[s])\n", line);
		return;
	}

	if(!app)
	{
		err_append("  ignoring line %d (command line without preceding application definition)\n", line);
		return;
	}

	/* check if fake key name contains a modifier (SHIFT+x, CTRL+x, ALT+x) */
	char mod = 0;
	{
	char* p = strchr(p2, '+');
	if(p)
	{
		if(!strncmp(p2, "SHIFT", 5))
			mod = VK_SHIFT;
		else if(!strncmp(p2, "CTRL", 4))
			mod = VK_CONTROL;
		else if(!strncmp(p2, "ALT", 3))
			mod = VK_MENU;
		else
		{
			err_append("  ignoring line %d (invalid key/modifier)\n", line);
			return;
		}
		p2 = p+1;
	}
	}

	/* check for repetition specifier */
	int reps = 1;
	{
	char* p = strchr(p2, '*');
	if(p)
	{
		reps = atoi(p+1);
		*p = '\0';				/* terminate string for comparison below */
	}
	}

	/* get (rest of) fake key name */
	char c = p2[0];	/* assume it's one char, doesn't need to be translated */
	if(strlen(p2) != 1)
	{
		int i;
		for(i = 0; i < NUM_VK; i++)
			if(!strcmp(p2, vk_str[i]))
			{
				c = vk[i];
				break;
			}

		/* not found in symbol table */
		if(i == NUM_VK)
		{
			err_append("  ignoring line %d (invalid key code)\n", line);
			return;
		}
	}

	if(app->num_cmds >= MAX_CMDS)
	{
		err_append("  warning: ignoring all bindings for app %u after line %d (there are too many)\n", (app-apps) / sizeof(struct APP), line);
		return;
	}

	int idx = app->num_cmds;
	app->rc_keys[idx] = p1;
	app->keys[idx] = c;
	app->allow_rep[idx] = allow_rep;
	app->modifier[idx] = mod;
	app->reps[idx] = reps;
	app->num_cmds++;

	p1 += strlen(p1)+1;

	cmds++;
}

static int read_cfg()
{
	FILE* f = fopen("config.txt", "r");
	if(!f)
	{
		err_append("could not open config.txt\n");
		return -1;
	}

	char s[256];

	while(fgets(s, sizeof(s), f))
	{
		line++;

		/* ignore comments and blank lines */
		if(s[0] == ';' || s[0] == '\n' || s[0] == '\r')
			continue;

		/* key definition ("0x%x %s", key code, key name)*/
		if(s[0] == '0' && s[1] == 'x')
			key_def(s);
		/* app header line */
		else if('0' <= s[0] && s[0] <= '9')
			app_header(s);
		/* key binding */
		else
			key_binding(s);
	}

	fclose(f);

	if(cmds == 0)
	{
		err_append("\n  no valid command bindings found in file.\n");
		return -1;
	}

	return err_occurred()? 1 : 0;
		// return warning if error string not empty,
		// but we succeeded.
}



/*
 * find a window, given class name.
 *
 * first check all top level windows, and then all their children
 */

static HWND wnd;
static char* class_name;

static int CALLBACK enum_func(HWND parent_wnd, LPARAM /* l */)
{
	wnd = FindWindowExA(parent_wnd, 0, class_name, 0);
	return (wnd)? 0 : 1;	/* stop if found (return false) */
}


static HWND find_wnd(char* _class_name)
{
	wnd = FindWindowA(_class_name, 0);
	if(wnd)
		return wnd;

	class_name = _class_name;
	EnumWindows(enum_func, 0);

	return wnd;
}


/*****************************************************************************
 * handle key events
 ****************************************************************************/


static int s;	/* socket */


/* send command to active app */
static int user_cmd(char* k, int reps)
{
	for(int i = 0; i < active->num_cmds; i++)
	{
		if(strcmp(k, active->rc_keys[i]))	/* does not match - next */
			continue;

		/* either key was pressed once (not held), or autorepeat is allowed */
		if(!reps || active->allow_rep[i])
		{
			/* send command several times, if so configured */
			for(int reps = 0; reps < active->reps[i]; reps++)
			{
				/*
				 * can't just send a VK_SHIFT WM_KEYDOWN - seems Windows
				 * maintains a shift state, which is queried by the app.
				 * we need to 'fake' the modifier globally, and reset it
				 * after sending the message.
				 */
				char mod = active->modifier[i];
				if(mod)
					keybd_event(mod, 0, 0, 0);
				PostMessageA(active->hwnd, WM_KEYDOWN, active->keys[i], 1);
				if(mod)
					keybd_event(mod, 0, KEYEVENTF_KEYUP, 0);
			}
		}

		return 1;	/* found it */
	}

	return 0;	/* no cmd bound */
}


static void builtin_cmd(char* k, int reps)
{
	// adjust global volume, if the app doesn't process "vol+" or "vol-"
	if(k[0] == 'v' && k[1] == 'o' && k[2] == 'l')	// +/- checked below
	{
		const long dv = 0x2800;
		DWORD vol;
		waveOutGetVolume(0, &vol);
		u16 vol1 = (u16)(vol & 0xffff), vol2 = (u16)(vol >> 16);
		if(k[3] == '+')
		{
			vol1 = addusw(vol1, dv);
			vol2 = addusw(vol2, dv);
		}
		else if(k[3] == '-')
		{
			vol1 = subusw(vol1, dv);
			vol2 = subusw(vol2, dv);
		}
		// otherwise, no-op
		vol = ((u32)vol2 << 16) | vol1;
		waveOutSetVolume(0, vol);
	}


	/* remaining builtin commands aren't autorepeatable */
	if(reps > 0)
		return;


	/*
	 * digits 1..9 - start / bring app to front
	 */
	if(('1' <= k[0] && k[0] <= '9') && k[1] == '\0')
	{
		active = &apps[k[0]-'0'];

		// already running? bring it to front
		if(active->hwnd)
		{
			ShowWindow(active->hwnd, SW_SHOW);
			SetForegroundWindow(active->hwnd);
		}
		// no - start it (if an app is bound to this key at all)
		else if(active->exe)
			/*
			 * returns when target app calls GetMessage();
			 * find_wnd called after next cmd is received
			 */
				WinExec(active->exe, SW_SHOWNORMAL);

		return;
	}

	/*
	 * close the active app
	 */
	if(!strcmp(k, "power"))
	{
		PostMessageA(active->hwnd, WM_CLOSE, 0, 0);
		active->hwnd = 0;

		return;
	}


	/*
	 * mute: toggle wave out volume on / off
	 */
	if(!strcmp(k, "mute"))
	{
		static DWORD vol;
		static bool is_muted;

		if(is_muted)
			waveOutSetVolume(0, vol);
		else
		{
			waveOutGetVolume(0, &vol);
			waveOutSetVolume(0, 0);
		}

		is_muted = !is_muted;

		return;
	}
}


static void* client_func(void*)
{
	for(;;)
	{
		char buf[100];
		if(recv(s, buf, sizeof(buf), 0) <= 0)
		{
			Sleep(0);	/* don't deadlock after an error */
			continue;
		}

		/*
		 * get app's window handle (may have been closed / new app started)
		 * this takes ~1 ms if the window isn't found or if it's a child
		 */
		active->hwnd = find_wnd(active->class_name);

		/* decode message */
		int code, reps;
		sscanf(buf+8, "%x %x", &code, &reps);
		char* k = 0;
		if(code < MAX_KEYS)
			k = key_strs[code];
debug_out("%x %s\n", code, k? k : "(not bound)");
		if(!k)
			continue;


		/* allow user cmd to override builtins */
		if(!user_cmd(k, reps))
			builtin_cmd(k, reps);
	}
}


void client_init()
{
	int ret = read_cfg();
	if(ret != 0)
		display_msg("IRRC problem", err);
	if(ret < 0)
		exit(1);

	s = (int)socket(AF_INET, SOCK_STREAM, 0);

	static struct sockaddr_in addr;
	addr.sin_family = AF_INET;
	addr.sin_port = htons(8765);
	addr.sin_addr.s_addr = inet_addr("127.0.0.1");
	if(connect(s, (struct sockaddr*)&addr, sizeof(struct sockaddr_in)))
	{
		display_msg("IRRC problem", "Could not connect to IR driver.\n");
		exit(1);
	}

	pthread_create(0, 0, client_func, 0);
}

/*Key press simulation. You can specify Alt, Ctrl or Shift modifiers. 
Mouse movement and click control. 
Execution of external applications. 
PostMessage API function access to send Windows type messages. 
Volume control. 
Computer log off, reboot or shutdown. 
Windows management. 
On Screen Display of system parameters or free text.*/
