/*
 * IRRC - decode
 * Copyright (c) 2003 Jan Wassenberg
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Contact info:
 *   Jan.Wassenberg@stud.uni-karlsruhe.de
 *   http://www.stud.uni-karlsruhe.de/~urkt/
 *
 */

/*
 * Notes:
 * - by popular demand, supports NEC ('Japanese') format
 *   as well as RC5.
 *
 * - decoding after the packet is complete (e.g. no activity for 3 ms)
 *   is tempting, but it'd require a timeout waiting for DCD to change,
 *   or we'd always be 1 packet behind => decode on the fly.
 */

#include <cmath>
#include <cstdio>

#include "irrc.h"

#include "adts.h"
#include "timer.h"


/*
 * rc5/rc6 decoder
 *
 * params:
 *   dt - time [s] since last transition
 *
 * error strategy:
 * - ignore very short transitions (may occur during a packet)
 * - treat transitions after a long delay as the start of a packet
 *   (last packet may have been interrupted, or we missed a transition)
 *
 * return 0 if signal was rejected (e.g. too short) - last_t will not be upated;
 *        1 otherwise.
 */
static int rc5_decode(double dt)
{
	/*
	 * on the fly bi-phase / manchester decode algorithm: see
	 * http://www.clearwater.com.au/pcm-9574/rc5/rc5.html
	 */
	static enum
	{
		WAITING,	/* waiting for start of next packet */
		START,		/* start of bit */
		MIDDLE		/* middle of bit */
	}
	state = WAITING;

	/* bit timings [s] (inclusive) */
	const double S_MIN =  444e-6, S_MAX = 1333e-6;
	const double L_MIN = 1334e-6, L_MAX = 2222e-6;

	/* first pulse of packet */
	if(dt > L_MAX || state == WAITING)
	{
		bit_buf.reset();

		/* HACK: force state change, emit '1' (start bit) */
		dt = (L_MIN+L_MAX)/2.;
	}

	/* too short: ignore it, don't update last_t */
	if(dt < S_MIN)
		return 0;

	/* short */
	if(dt <= S_MAX)
	{
		if(state != START)
			state = START;
		else
		{
			state = MIDDLE;
			bit_buf.add(0);
		}
	}
	/* long */
	else
	{
		state = MIDDLE;
		bit_buf.add(1);
	}

	/*
	 * packet complete?
	 * (we make sure the last transition is consumed)
	 */
	if(bit_buf.len == 14 && !(bit_buf.cur == 1 && state == MIDDLE))
	{
		state = WAITING;

		uint code   = bit_buf.extract(6);
		uint addr   = bit_buf.extract(5);
		uint toggle = bit_buf.extract(1);
		uint start  = bit_buf.extract(2);

		/* only react to stuff sent to TV1 */
		if(addr != 0)
			return 1;

		/* RC6 mess: code[6] = !start[0] */
		code |= (start ^ 3) << 6;

		/* how long has key been pressed? */
		static uint last_toggle = 2;

		static uint reps;
		reps++;						/* first time: set to 0 below since last_toggle = impossible_val */
		if(toggle != last_toggle)
		{
			reps = 0; 
			last_toggle = toggle;
		}

		server_send(code, reps);
	}

	return 1;	/* success, update last_t */
}


// NEC decoder
//
// params:
//   dt - time [s] since last transition
//
// error strategy:
// - ignore very short transitions (may occur during a packet)
// - treat transitions after a long delay as the start of a packet
//   (last packet may have been interrupted, or we missed a transition)
//
// return 0 if signal was rejected (e.g. too short) - last_t will not be upated;
//        1 otherwise.
static int nec_decode(double dt)
{
	static enum
	{
		WAITING,
		AGC_PULSE,
		AGC_SPACE,
		PULSE,
		SPACE__,
		REPEAT_PULSE
	}
	state = WAITING;

	// used to verify a repeat packet (they come at 110 ms intervals)
	static double repeat_packet_dt;
	static int reps;

	// can't use bit_buf - NEC sends in reverse order
	static unsigned long bits;	// will hold 32 bits
	static int nbits;

	// too short - probably noise. ignore, don't update last_t
	if(dt < 0.30e-3)
		return 0;
	// either start of packet anyway, or we were still waiting on pulses
    // that didn't come (probably missed them). reset.
	if(dt > 15e-3)
		state = WAITING;

	// decode FSM
	// protocol:
	//  _________      _   _     _
	// |         |____| |_| |___| |_ ...
	//      a      b   c 0 c  1  c 0
	// a: 9ms AGC burst
	// b: 4.5ms space
	// c: 560s pulse
	// followed by a 560s space for a 0 bit, or 1.69 ms for a 1
	// packet format: LSB first, 8 bit address, its 1-s complement,
	//   then command and its complement (for error detection).
	//
	// repeat packets: only a, b(with length 2.25 ms), c; sent every 110 ms
	// (time since start of data packet or last repeat packet)
	//
	// we're called after a DCD leading/trailing flank. state indicates
	// which part of the pulse train has just ended; we check if its length
	// was correct via dt (note: protocol seems to have HUGE tolerances
	// here, up to 40%).

	switch(state)
	{
	case WAITING:
		// most likely a bogus pulse (turnaround isn't that fast)
		if(dt < 10e-3)
			return 0;

		state = AGC_PULSE;

		// reset
		bits = 0; nbits = 0;
		repeat_packet_dt = dt;
			// only used if this turns out to be a repeat packet
		break;

	case AGC_PULSE:
		if(6e-3 < dt && dt < 14e-3)	// nominal: 9 ms
			state = AGC_SPACE;
		else
			goto reset;
		break;

	case AGC_SPACE:
		if(2e-3 < dt && dt < 2.5e-3)	// nominal: 2.25 ms
			state = REPEAT_PULSE;
		else if(4e-3 < dt && dt < 5e-3)	// nominal: 4.5 ms
			state = PULSE;
		else
			goto reset;
		break;

	case PULSE:
	case REPEAT_PULSE:
		if(!(0.40e-3 < dt && dt < 0.9e-3))	// nominal: 560 s
			goto reset;
		if(state == PULSE)
			state = SPACE__;
		// repeat packet
		else
		{
			// repeat packets (length: ~12 ms) should at 110 ms intervals
			// min time difference = 30 ms for first repeat packet
			// (110 ms spacing - data packet length of 68 ms)
			if(30e-3 < repeat_packet_dt && repeat_packet_dt < 110e-3)
			{
				reps++;
				goto repeat;	// state set to WAITING there
			}
			else
				goto reset;
		}
		break;

	case SPACE__:
		state = PULSE;
		// note: if less than 300 s, it got filtered out above
		if(dt < 1.125e-3)	// nominal: 560 s
		{
			bits >>= 1;
			nbits++;
		}
		else if(dt < 2.5e-3)	// nominal: 1.69 ms
		{
			bits >>= 1;
			bits |= 0x80000000;
			nbits++;
		}
		else
			goto reset;
	}

	// packet not yet complete, we're done for now
	if(nbits != 32)
		return 1;

	//
	// decode packet
	//
	static uint adr, cmd;
	uint _adr, _cmd;	// their 1-s complement (also sent)

	 adr = bits & 0xff; bits >>= 8;
	_adr = bits & 0xff; bits >>= 8;
	 cmd = bits & 0xff; bits >>= 8;
	_cmd = bits & 0xff; bits >>= 8;
	nbits = 0;

	// checksum failed
	if(cmd+_cmd != 0xff)
		goto reset;

	reps = 0;

repeat:
	server_send(cmd, reps);

	state = WAITING;
	return 1;

reset:
	state = WAITING;
	return 1;
}


void decode(int dcd)
{
	static double last_t;
	double t = get_time();

/*
static int last_dcd = 10;
if(last_dcd == dcd)
{
	OutputDebugString("\nMISSED TRANSITION\n\n");
}
last_dcd = dcd;
*/

	if(rc5_decode(t - last_t))
		last_t = t;
//	if(nec_decode(t - last_t))
//		last_t = t;
}
