/*
 * IRRC
 * Copyright (c) 2003 Jan Wassenberg
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Contact info:
 *   Jan.Wassenberg@stud.uni-karlsruhe.de
 *   http://www.stud.uni-karlsruhe.de/~urkt/
 *
 */

#include <stdlib.h>

#include "posix.h"

#include "sysdep/win/win_internal.h"

#include "irrc.h"


static int fd = -1;

#ifdef _WIN32
static HANDLE h;
#endif

static int input;


/* HW not found: display message, wait for input */
static void* hw_msg_func(void* /* param */)
{
	static const char* msg = "Could not find receiver hardware.\n"
							 "\n"
							 "Does it support autodetect?\n"
							 "  if so, plug it in;\n"
#ifdef _WIN32
							 "  otherwise, plug it into COM1 and choose 'ok'.\n"
							 "\n"
							 "'cancel' exits the program.\n";
#else
							 "  otherwise, plug it into tty0 and press any key.\n"
							 "'x' exits the program.\n";
#endif

#ifdef _WIN32
	input = MessageBoxA(0, msg, "IRRC", MB_OKCANCEL|MB_SETFOREGROUND);
#else
	printf("%s", msg);

	/* non-canonical mode, no echo */
	struct termios tio;
	tcgetattr(0, &tio);
	tio.c_lflag &= ~(ICANON|ECHO);
	tcsetattr(0, TCSANOW, &tio);

	char c;
	read(0, &c, 1);
	input = c;
#endif

	return 0;
}


/*
 * scan ports, check if HW is attached
 * (if it supports autodetect, CTS and RTS are connected)
 */
static bool hw_is_present()
{
	/*
	 * check in reverse order: in case HW doesn't support autodetect,
	 * leave 1st port open (see below).
	 */
	for(int i = 1; i >= 0; i--)
	{
		static char port[] = "/dev/tty?";
		port[8] = (char)('0'+i);

		if(fd != -1)
			close(fd);
		fd = open(port, O_RDWR | O_NONBLOCK);	/* nonblock => open returns w/o DCD */
		if(fd == -1)
			continue;

#ifdef _WIN32
		h = (HANDLE)((char*)0 + _get_osfhandle(fd));
		SetCommMask(h, EV_CTS|EV_RLSD);
#endif

		/*
		 * check if HW is attached to this port.
		 *
		 * RTS line powers the receiver.
		 * cts0 := CTS with RTS cleared; cts1 := CTS with RTS asserted.
		 * (cts0 == 0 && cts1 == 1) ==> (CTS <==> RTS) ==> device found.
		 * 1 ms settle time.
		 */

#ifdef _WIN32
		DWORD status;

		int rts0 = CLRRTS;
		EscapeCommFunction(h, rts0);
		Sleep(1);
		GetCommModemStatus(h, &status);
		int cts0 = (status & MS_CTS_ON) != 0;

		int rts1 = SETRTS;
		EscapeCommFunction(h, rts1);
		Sleep(1);
		GetCommModemStatus(h, &status);
		int cts1 = (status & MS_CTS_ON) != 0;
#else
		int sigs;

		int rts0 = 0;
		ioctl(fd, TIOCMBIS, &rts0);
		usleep(1000);
		ioctl(fd, TIOCMGET, &sigs);
		int cts0 = (sigs & TIOCM_CTS) != 0;

		int rts1 = TIOCM_RTS;
		ioctl(fd, TIOCMBIS, &rts1);
		usleep(1000);
		ioctl(fd, TIOCMGET, &sigs);
		int cts1 = (sigs & TIOCM_CTS) != 0;
#endif

		/* found it */
		if(cts0 == 0 && cts1 == 1)
			return true;
	}

	return false;
}


/* wait until HW is found, or the user chooses to continue or exit. */
void hw_wait_present()
{
	if(hw_is_present())
		return;

	/* show message (thread checks for input) */
	pthread_t msg_thread;
	pthread_create(&msg_thread, 0, hw_msg_func, 0);

	/* wait for input or device to be attached */
	input = 0;
	while(input == 0 && !hw_is_present())
		;

#ifdef _WIN32
	if(input == IDCANCEL)
#else
	if(input == 'x')
#endif
		exit(1);

	/* get rid of input thread */
#ifdef _WIN32
	HWND dlg = FindWindowA(0, "IRRC");
	PostMessageA(dlg, WM_CLOSE, 0, 0);	/* closes thread */
#else
	pthread_cancel(msg_thread);
#endif
}


int hw_wait_dcd_change()
{
	/* wait for DCD or CTS to change */
#ifdef _WIN32
	DWORD mask;
	WaitCommEvent(h, &mask, 0);		/* comm mask set at init */
#else
	int mask = TIOCM_CD|TIOCM_CTS;
	ioctl(fd, TIOCMIWAIT, &mask);
#endif

	/* if CTS went low, HW was disconnected; wait for reconnect */
#ifdef _WIN32
	DWORD status;
	GetCommModemStatus(h, &status);
	if((mask & EV_CTS) && !(status & MS_CTS_ON))
		hw_wait_present();

	int dcd = !!(status & MS_RLSD_ON);

#else
	int sigs;
	ioctl(fd, TIOCMGET, &sigs);
	if((mask & TIOCM_CTS) && !(sigs & TIOCM_CTS))
		hw_wait_present();

	int dcd = !!(sigs & TIOCM_CTS);
#endif

	return dcd;
}
